# Copyright (c) The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg

TEST_CONF = None

identity_group = cfg.OptGroup(name='identity')
identity_options = [
    cfg.StrOpt(
        'auth_url',
        default='http://localhost/identity/v3',
        help='Keystone endpoint',
    ),
    cfg.StrOpt('username', default='admin', help='Keystone username'),
    cfg.StrOpt(
        'password',
        default='secretadmin',
        help='Password used with Keystone username',
    ),
    cfg.StrOpt(
        'project_name',
        default='admin',
        help='Name of project, used by the given username',
    ),
    cfg.StrOpt(
        'user_domain_name',
        default='Default',
        help='Name of domain, used by the given username',
    ),
    cfg.StrOpt(
        'project_domain_name',
        default='Default',
        help='Name of domain, used by the given project',
    ),
]


def setup_config(config_file=''):
    global TEST_CONF
    TEST_CONF = cfg.ConfigOpts()

    TEST_CONF.register_group(identity_group)
    TEST_CONF.register_opts(identity_options, group=identity_group)

    config_to_load = []
    local_config = './etc/castellan/castellan-functional.conf'
    main_config = '/etc/castellan/castellan-functional.conf'
    if os.path.isfile(config_file):
        config_to_load.append(config_file)
    elif os.path.isfile(local_config):
        config_to_load.append(local_config)
    elif os.path.isfile(main_config):
        config_to_load.append(main_config)

    TEST_CONF(
        [],  # Required to load an anonymous configuration
        default_config_files=config_to_load,
    )


def get_config():
    if not TEST_CONF:
        setup_config()
    return TEST_CONF


def list_opts():
    yield identity_group.name, identity_options



# Copyright (c) 2016 IBM
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common utilities for Castellan.
"""

from typing import Any

from oslo_config import cfg
from oslo_log import log as logging

from castellan.common.credentials import credential
from castellan.common.credentials import keystone_password
from castellan.common.credentials import keystone_token
from castellan.common.credentials import password
from castellan.common.credentials import token
from castellan.common import exception


LOG = logging.getLogger(__name__)

credential_opts = [
    # auth_type opt
    cfg.StrOpt(
        'auth_type',
        choices=('token', 'password', 'keystone_token', 'keystone_password'),
        help="The type of authentication credential to create. "
        "Required if no context is passed to the credential factory.",
    ),
    # token opt
    cfg.StrOpt(
        'token',
        secret=True,
        help="Token for authentication. Required for 'token' and "
        "'keystone_token' auth_type if no context is passed to the "
        "credential factory.",
    ),
    # password opts
    cfg.StrOpt(
        'username',
        help="Username for authentication. Required for 'password' "
        "auth_type. Optional for the 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'password',
        secret=True,
        help="Password for authentication. Required for 'password' and "
        "'keystone_password' auth_type.",
    ),
    # keystone credential opts
    cfg.URIOpt('auth_url', help="Use this endpoint to connect to Keystone."),
    cfg.StrOpt(
        'user_id',
        help="User ID for authentication. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'user_domain_id',
        help="User's domain ID for authentication. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'user_domain_name',
        help="User's domain name for authentication. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'trust_id',
        help="Trust ID for trust scoping. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'domain_id',
        help="Domain ID for domain scoping. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'domain_name',
        help="Domain name for domain scoping. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'project_id',
        help="Project ID for project scoping. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'project_name',
        help="Project name for project scoping. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'project_domain_id',
        help="Project's domain ID for project. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.StrOpt(
        'project_domain_name',
        help="Project's domain name for project. Optional for "
        "'keystone_token' and 'keystone_password' auth_type.",
    ),
    cfg.BoolOpt(
        'reauthenticate',
        default=True,
        help="Allow fetching a new token if the current one is "
        "going to expire. Optional for 'keystone_token' and "
        "'keystone_password' auth_type.",
    ),
]

OPT_GROUP = 'key_manager'


def credential_factory(
    conf: cfg.ConfigOpts | None = None, context: Any = None
) -> credential.Credential:
    """This function provides a factory for credentials.

    It is used to create an appropriare credential object
    from a passed configuration. This should be called before
    making any calls to a key manager.

    :param conf: Configuration file which this factory method uses
    to generate a credential object. Note: In the future it will
    become a required field.
    :param context: Context used for authentication. It can be used
    in conjunction with the configuration file. If no conf is passed,
    then the context object will be converted to a KeystoneToken and
    returned. If a conf is passed then only the 'token' is grabbed from
    the context for the authentication types that require a token.
    :returns: A credential object used for authenticating with the
    Castellan key manager. Type of credential returned depends on
    config and/or context passed.
    """
    if conf:
        conf.register_opts(credential_opts, group=OPT_GROUP)

        if conf.key_manager.auth_type == 'token':
            if conf.key_manager.token:
                auth_token = conf.key_manager.token
            elif context:
                auth_token = context.auth_token
            else:
                raise exception.InsufficientCredentialDataError()

            return token.Token(auth_token)

        elif conf.key_manager.auth_type == 'password':
            return password.Password(
                conf.key_manager.username, conf.key_manager.password
            )

        elif conf.key_manager.auth_type == 'keystone_password':
            return keystone_password.KeystonePassword(
                conf.key_manager.password,
                auth_url=conf.key_manager.auth_url,
                username=conf.key_manager.username,
                user_id=conf.key_manager.user_id,
                user_domain_id=conf.key_manager.user_domain_id,
                user_domain_name=conf.key_manager.user_domain_name,
                trust_id=conf.key_manager.trust_id,
                domain_id=conf.key_manager.domain_id,
                domain_name=conf.key_manager.domain_name,
                project_id=conf.key_manager.project_id,
                project_name=conf.key_manager.project_name,
                project_domain_id=conf.key_manager.project_domain_id,
                project_domain_name=conf.key_manager.project_domain_name,
                reauthenticate=conf.key_manager.reauthenticate,
            )

        elif conf.key_manager.auth_type == 'keystone_token':
            if conf.key_manager.token:
                auth_token = conf.key_manager.token
            elif context:
                auth_token = context.auth_token
            else:
                raise exception.InsufficientCredentialDataError()

            return keystone_token.KeystoneToken(
                auth_token,
                auth_url=conf.key_manager.auth_url,
                trust_id=conf.key_manager.trust_id,
                domain_id=conf.key_manager.domain_id,
                domain_name=conf.key_manager.domain_name,
                project_id=conf.key_manager.project_id,
                project_name=conf.key_manager.project_name,
                project_domain_id=conf.key_manager.project_domain_id,
                project_domain_name=conf.key_manager.project_domain_name,
                reauthenticate=conf.key_manager.reauthenticate,
            )

        else:
            LOG.error("Invalid auth_type specified.")
            raise exception.AuthTypeInvalidError(
                type=conf.key_manager.auth_type
            )

    # for compatibility between _TokenData and RequestContext
    project_id: str | None = None
    if hasattr(context, 'tenant') and context.tenant:
        project_id = context.tenant
    elif hasattr(context, 'project_id') and context.project_id:
        project_id = context.project_id

    return keystone_token.KeystoneToken(
        context.auth_token, project_id=project_id
    )



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Key manager implementation that raises NotImplementedError
"""

from typing import Any

from oslo_config import cfg

from castellan.common.objects import managed_object
from castellan.key_manager import key_manager
from castellan.key_manager.key_manager import Context


class NotImplementedKeyManager(key_manager.KeyManager):
    """Key Manager that raises NotImplementedError for all operations."""

    def __init__(self, configuration: cfg.ConfigOpts | None = None) -> None:
        pass

    def create_key(
        self,
        context: Context | None,
        algorithm: str = 'AES',
        length: int = 256,
        expiration: str | None = None,
        name: str | None = None,
    ) -> str:
        raise NotImplementedError()

    def create_key_pair(
        self,
        context: Context | None,
        algorithm: str = 'AES',
        length: int = 256,
        expiration: str | None = None,
        name: str | None = None,
    ) -> tuple[str, str]:
        raise NotImplementedError()

    def store(
        self,
        context: Context | None,
        managed_object: managed_object.ManagedObject,
        expiration: str | None = None,
    ) -> str:
        raise NotImplementedError()

    # TODO(stephenfin): This is not defined on the base class nor on other
    # in-tree key managers. Can we remove it?
    def copy(
        self,
        context: Context | None,
        managed_object_id: str,
        **kwargs: Any,
    ) -> Any:
        raise NotImplementedError()

    def get(
        self,
        context: Context | None,
        managed_object_id: str,
        metadata_only: bool = False,
    ) -> managed_object.ManagedObject:
        raise NotImplementedError()

    def list(
        self,
        context: Context | None,
        object_type: type[managed_object.ManagedObject] | None = None,
        metadata_only: bool = False,
    ) -> list[managed_object.ManagedObject]:
        raise NotImplementedError()

    def delete(
        self,
        context: Context | None,
        managed_object_id: str,
        force: bool = False,
    ) -> None:
        raise NotImplementedError()

    def add_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        raise NotImplementedError()

    def remove_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        raise NotImplementedError()



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
X509 Class

This module defines the X509 class, used to represent X.509 certificates.
"""

from castellan.common.objects import certificate


class X509(certificate.Certificate):
    """This class represents X.509 certificates."""

    @classmethod
    def managed_type(cls) -> str:
        return "certificate"

    @property
    def format(self) -> str:
        return "X.509"

    def get_encoded(self) -> bytes | None:
        return self._data

    def __eq__(self, other: object) -> bool:
        if isinstance(other, X509):
            return self._data == other._data
        else:
            return False

    def __ne__(self, other: object) -> bool:
        result = self.__eq__(other)
        return not result



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the key manager.
"""

from oslo_config import cfg
from oslo_config import fixture

from castellan import key_manager
from castellan.key_manager import barbican_key_manager
from castellan.tests import base

CONF = cfg.CONF


class KeyManagerTestCase(base.TestCase):
    def _create_key_manager(self):
        raise NotImplementedError()

    def setUp(self):
        super().setUp()

        self.conf = self.useFixture(fixture.Config()).conf

        self.key_mgr = self._create_key_manager()


class DefaultKeyManagerImplTestCase(KeyManagerTestCase):
    def _create_key_manager(self):
        return key_manager.API(self.conf)

    def test_default_key_manager(self):
        self.assertEqual("barbican", self.conf.key_manager.backend)
        self.assertIsNotNone(self.key_mgr)
        self.assertIsInstance(
            self.key_mgr, barbican_key_manager.BarbicanKeyManager
        )



# Copyright (c) 2015 IBM
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the keystone password credential
"""

from castellan.common.credentials import keystone_password
from castellan.tests import base


class KeystonePasswordTestCase(base.TestCase):
    def _create_ks_password_credential(self):
        return keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

    def setUp(self):
        self.password = "Pa$$w0rd1"
        self.username = "admin"
        self.user_id = "1adb2391c009443aa5224b316d4a06ae"
        self.user_domain_id = "default"
        self.user_domain_name = "default"
        self.trust_id = "14b38a8296f144148138466ce9280940"
        self.domain_id = "default"
        self.domain_name = "default"
        self.project_id = "1099302ec608486f9879ba2466c60720"
        self.project_name = "demo"
        self.project_domain_id = "default"
        self.project_domain_name = "default"
        self.reauthenticate = True

        self.ks_password_credential = self._create_ks_password_credential()

        super().setUp()

    def test_get_password(self):
        self.assertEqual(self.password, self.ks_password_credential.password)

    def test_get_username(self):
        self.assertEqual(self.username, self.ks_password_credential.username)

    def test_get_user_id(self):
        self.assertEqual(self.user_id, self.ks_password_credential.user_id)

    def test_get_user_domain_id(self):
        self.assertEqual(
            self.user_domain_id, self.ks_password_credential.user_domain_id
        )

    def test_get_user_domain_name(self):
        self.assertEqual(
            self.user_domain_name, self.ks_password_credential.user_domain_name
        )

    def test_get_trust_id(self):
        self.assertEqual(self.trust_id, self.ks_password_credential.trust_id)

    def test_get_domain_id(self):
        self.assertEqual(self.domain_id, self.ks_password_credential.domain_id)

    def test_get_domain_name(self):
        self.assertEqual(
            self.domain_name, self.ks_password_credential.domain_name
        )

    def test_get_project_id(self):
        self.assertEqual(
            self.project_id, self.ks_password_credential.project_id
        )

    def test_get_project_name(self):
        self.assertEqual(
            self.project_name, self.ks_password_credential.project_name
        )

    def test_get_project_domain_id(self):
        self.assertEqual(
            self.project_domain_id,
            self.ks_password_credential.project_domain_id,
        )

    def test_get_project_domain_name(self):
        self.assertEqual(
            self.project_domain_name,
            self.ks_password_credential.project_domain_name,
        )

    def test_get_reauthenticate(self):
        self.assertEqual(
            self.reauthenticate, self.ks_password_credential.reauthenticate
        )

    def test___eq__(self):
        self.assertTrue(
            self.ks_password_credential == self.ks_password_credential
        )
        self.assertTrue(
            self.ks_password_credential is self.ks_password_credential
        )

        self.assertFalse(self.ks_password_credential is None)
        self.assertFalse(None == self.ks_password_credential)  # noqa: E711

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )
        self.assertTrue(
            self.ks_password_credential == other_ks_password_credential
        )
        self.assertFalse(
            self.ks_password_credential is other_ks_password_credential
        )

    def test___ne___none(self):
        self.assertTrue(self.ks_password_credential is not None)
        self.assertTrue(None != self.ks_password_credential)  # noqa: E711

    def test___ne___password(self):
        other_password = "wheresmyCat??"

        other_ks_password_credential = keystone_password.KeystonePassword(
            other_password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___username(self):
        other_username = "service"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=other_username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___user_id(self):
        other_user_id = "service"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=other_user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___user_domain_id(self):
        other_user_domain_id = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=other_user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___user_domain_name(self):
        other_user_domain_name = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.domain_id,
            user_domain_name=other_user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___trust_id(self):
        other_trust_id = "00000000000000"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=other_trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___domain_id(self):
        other_domain_id = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=other_domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___domain_name(self):
        other_domain_name = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=other_domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___project_id(self):
        other_project_id = "00000000000000"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=other_project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___project_name(self):
        other_project_name = "proj0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=other_project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___project_domain_id(self):
        other_project_domain_id = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=other_project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___project_domain_name(self):
        other_project_domain_name = "domain0"

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=other_project_domain_name,
            reauthenticate=self.reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )

    def test___ne___reauthenticate(self):
        other_reauthenticate = False

        other_ks_password_credential = keystone_password.KeystonePassword(
            self.password,
            username=self.username,
            user_id=self.user_id,
            user_domain_id=self.user_domain_id,
            user_domain_name=self.user_domain_name,
            trust_id=self.trust_id,
            domain_id=self.domain_id,
            domain_name=self.domain_name,
            project_id=self.project_id,
            project_name=self.project_name,
            project_domain_id=self.project_domain_id,
            project_domain_name=self.project_domain_name,
            reauthenticate=other_reauthenticate,
        )

        self.assertTrue(
            self.ks_password_credential != other_ks_password_credential
        )



# Copyright (c) The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Key manager implementation for Barbican
"""

from __future__ import annotations

import builtins
import calendar
import time
from typing import Any, TypeAlias, cast

from barbicanclient import client as barbican_client_import
from barbicanclient import exceptions as barbican_exceptions
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization
from cryptography import x509 as cryptography_x509
from keystoneauth1 import identity
from keystoneauth1 import loading
from keystoneauth1 import plugin
from keystoneauth1 import service_token
from keystoneauth1 import session
from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import timeutils

from castellan.common.credentials import keystone_password
from castellan.common.credentials import keystone_token
from castellan.common import exception
from castellan.common.objects import key as key_base_class
from castellan.common.objects import managed_object
from castellan.common.objects import opaque_data as op_data
from castellan.i18n import _
from castellan.key_manager import key_manager
from castellan.key_manager.key_manager import Context

_barbican_opts = [
    cfg.URIOpt(
        'barbican_endpoint',
        help='Use this endpoint to connect to Barbican, for example: '
        '"http://localhost:9311/"',
    ),
    cfg.StrOpt(
        'barbican_api_version',
        help='Version of the Barbican API, for example: "v1"',
    ),
    cfg.URIOpt(
        'auth_endpoint',
        default='http://localhost/identity/v3',
        deprecated_name='auth_url',
        deprecated_group='key_manager',
        help='Use this endpoint to connect to Keystone',
    ),
    cfg.IntOpt(
        'retry_delay',
        default=1,
        help='Number of seconds to wait before retrying poll for key '
        'creation completion',
    ),
    cfg.IntOpt(
        'number_of_retries',
        default=60,
        help='Number of times to retry poll for key creation completion',
    ),
    cfg.BoolOpt(
        'verify_ssl',
        default=True,
        help='Specifies if insecure TLS (https) requests. If False, '
        'the server\'s certificate will not be validated, if '
        'True, we can set the verify_ssl_path config meanwhile.',
    ),
    cfg.StrOpt(
        'verify_ssl_path',
        default=None,
        help='A path to a bundle or CA certs to check against, or '
        'None for requests to attempt to locate and use '
        'certificates which verify_ssh is True. If verify_ssl '
        'is False, this is ignored.',
    ),
    cfg.StrOpt(
        'barbican_endpoint_type',
        default='public',
        choices=['public', 'internal', 'admin'],
        help='Specifies the type of endpoint.',
    ),
    cfg.StrOpt(
        'barbican_region_name',
        default=None,
        help='Specifies the region of the chosen endpoint.',
    ),
    cfg.BoolOpt(
        'send_service_user_token',
        default=False,
        help="""
When True, if sending a user token to a REST API, also send a service token.

Nova often reuses the user token provided to the nova-api to talk to other REST
APIs, such as Cinder, Glance and Neutron. It is possible that while the user
token was valid when the request was made to Nova, the token may expire before
it reaches the other service. To avoid any failures, and to make it clear it is
Nova calling the service on the user's behalf, we include a service token along
with the user token. Should the user's token have expired, a valid service
token ensures the REST API request will still be accepted by the keystone
middleware.
""",
    ),
]

_BARBICAN_OPT_GROUP = 'barbican'
_BARBICAN_SERVICE_USER_OPT_GROUP = 'barbican_service_user'

LOG = logging.getLogger(__name__)


Auth: TypeAlias = (
    identity.Token
    | identity.Password
    | service_token.ServiceTokenAuthWrapper
    | plugin.BaseAuthPlugin
)


class BarbicanKeyManager(key_manager.KeyManager):
    """Key Manager Interface that wraps the Barbican client API."""

    conf: cfg.ConfigOpts
    _verify: str | bool

    def __init__(self, configuration: cfg.ConfigOpts) -> None:
        self.conf = configuration
        self.conf.register_opts(_barbican_opts, group=_BARBICAN_OPT_GROUP)
        loading.register_session_conf_options(self.conf, _BARBICAN_OPT_GROUP)

        loading.register_session_conf_options(
            self.conf, _BARBICAN_SERVICE_USER_OPT_GROUP
        )
        loading.register_auth_conf_options(
            self.conf, _BARBICAN_SERVICE_USER_OPT_GROUP
        )
        verify_ssl = self.conf.barbican.verify_ssl
        verify_ssl_path = self.conf.barbican.verify_ssl_path
        self._verify = verify_ssl and (verify_ssl_path or verify_ssl)

    def _get_barbican_client(
        self, context: Context | None
    ) -> barbican_client_import.Client:
        """Creates a client to connect to the Barbican service.

        :param context: the user context for authentication
        :return: a Barbican Client object
        :raises Forbidden: if the context is None
        :raises KeyManagerError: if context is missing tenant or tenant is
            None or error occurs while creating client
        """

        # Confirm context is provided, if not raise forbidden
        if not context:
            msg = _("User is not authorized to use key manager.")
            LOG.error(msg)
            raise exception.Forbidden(msg)

        try:
            auth = self._get_keystone_auth(context)
            sess = session.Session(auth=auth, verify=self._verify)

            endpoint = self._get_barbican_endpoint(auth, sess)
            return barbican_client_import.Client(
                version=self.conf.barbican.barbican_api_version,
                session=sess,
                endpoint=endpoint,
            )

        # TODO(pbourke): more fine grained exception handling - we are eating
        # tracebacks here
        except Exception as e:
            LOG.error("Error creating Barbican client: %s", e)
            raise exception.KeyManagerError(reason=e)

    def _get_keystone_auth(self, context: Context) -> Auth:
        auth: Auth

        if isinstance(context, keystone_password.KeystonePassword):
            if not context.auth_url:
                # this should never happen at runtime
                msg = _('missing auth_url')
                LOG.error(msg)
                raise exception.Forbidden(reason=msg)

            auth = identity.Password(
                auth_url=context.auth_url,
                username=context.username,
                password=context.password,
                user_id=context.user_id,
                user_domain_id=context.user_domain_id,
                user_domain_name=context.user_domain_name,
                trust_id=context.trust_id,
                domain_id=context.domain_id,
                domain_name=context.domain_name,
                project_id=context.project_id,
                project_name=context.project_name,
                project_domain_id=context.project_domain_id,
                project_domain_name=context.project_domain_name,
                reauthenticate=context.reauthenticate,
            )
        elif isinstance(context, keystone_token.KeystoneToken):
            if not context.auth_url:
                # this should never happen at runtime
                msg = _('missing auth_url')
                LOG.error(msg)
                raise exception.Forbidden(reason=msg)

            auth = identity.Token(
                auth_url=context.auth_url,
                token=context.token,
                trust_id=context.trust_id,
                domain_id=context.domain_id,
                domain_name=context.domain_name,
                project_id=context.project_id,
                project_name=context.project_name,
                project_domain_id=context.project_domain_id,
                project_domain_name=context.project_domain_name,
                reauthenticate=context.reauthenticate,
            )
        # this will be kept for oslo.context compatibility until
        # projects begin to use utils.credential_factory
        elif isinstance(context, oslo_context.RequestContext):
            # this is set by some projects' RequestContext subclasses
            if hasattr(context, 'get_auth_plugin'):
                auth = context.get_auth_plugin()
            else:
                if not context.auth_token:
                    # this should never happen at runtime
                    msg = _('missing auth_token')
                    LOG.error(msg)
                    raise exception.Forbidden(reason=msg)

                auth = identity.Token(
                    auth_url=self.conf.barbican.auth_endpoint,
                    token=context.auth_token,
                    project_id=context.project_id,
                    project_name=context.project_name,
                    project_domain_id=context.project_domain_id,
                    project_domain_name=context.project_domain_name,
                    domain_id=context.domain_id,
                    domain_name=context.domain_name,
                    system_scope=context.system_scope,
                )
        else:
            msg = _(
                "context must be of type KeystonePassword, "
                "KeystoneToken, or RequestContext."
            )
            LOG.error(msg)
            raise exception.Forbidden(reason=msg)

        if self.conf.barbican.send_service_user_token:
            service_auth = loading.load_auth_from_conf_options(
                self.conf, group=_BARBICAN_SERVICE_USER_OPT_GROUP
            )
            if service_auth is None:
                msg = _("missing configuration for service user")
                LOG.error(msg)
                raise exception.Forbidden(reason=msg)

            auth = service_token.ServiceTokenAuthWrapper(
                user_auth=auth,
                service_auth=service_auth,
            )

        return auth

    def _get_barbican_endpoint(
        self, auth: Auth, sess: session.Session
    ) -> str | None:
        if self.conf.barbican.barbican_endpoint:
            return cast(str, self.conf.barbican.barbican_endpoint)

        if hasattr(auth, 'service_catalog'):
            endpoint_data = auth.service_catalog.endpoint_data_for(
                service_type='key-manager',
                interface=self.conf.barbican.barbican_endpoint_type,
                region_name=self.conf.barbican.barbican_region_name,
            )
            return endpoint_data.url  # type: ignore[no-any-return]

        return auth.get_endpoint(
            sess,
            service_type='key-manager',
            interface=self.conf.barbican.barbican_endpoint_type,
            region_name=self.conf.barbican.barbican_region_name,
        )

    def _delete_order(
        self, client: barbican_client_import.Client, order_ref: str
    ) -> None:
        try:
            client.orders.delete(order_ref)
        except Exception as e:
            LOG.warning(
                "Failed to delete temporary order %s: %s", order_ref, e
            )

    def _delete_container(
        self, client: barbican_client_import.Client, container_ref: str
    ) -> None:
        try:
            client.containers.delete(container_ref)
        except Exception as e:
            LOG.warning(
                "Failed to delete temporary container %s: %s", container_ref, e
            )

    def create_key(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> str:
        """Creates a symmetric key.

        :param context: contains information of the user and the environment
            for the request
        :param algorithm: the algorithm associated with the secret
        :param length: the bit length of the secret
        :param name: the name of the key
        :param expiration: the date the key will expire
        :return: the UUID of the new key
        :raises KeyManagerError: if key creation fails
        """
        barbican_client = self._get_barbican_client(context)

        try:
            key_order = barbican_client.orders.create_key(
                name=name,
                algorithm=algorithm,
                bit_length=length,
                expiration=expiration,
            )
            order_ref = key_order.submit()
            order = self._get_active_order(barbican_client, order_ref)
            secret_ref = self._retrieve_secret_uuid(order.secret_ref)
            self._delete_order(barbican_client, order_ref)
            return secret_ref
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error creating key: %s", e)
            raise exception.KeyManagerError(reason=e)

    def create_key_pair(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> tuple[str, str]:
        """Creates an asymmetric key pair.

        :param context: contains information of the user and the environment
            for the request
        :param algorithm: the algorithm associated with the secret
        :param length: the bit length of the secret
        :param name: the name of the key
        :param expiration: the date the key will expire
        :return: the UUIDs of the new key, in the order (private, public)
        :raises NotImplementedError: until implemented
        :raises KeyManagerError: if key pair creation fails
        """
        barbican_client = self._get_barbican_client(context)

        try:
            key_pair_order = barbican_client.orders.create_asymmetric(
                algorithm=algorithm,
                bit_length=length,
                name=name,
                expiration=expiration,
            )

            order_ref = key_pair_order.submit()
            order = self._get_active_order(barbican_client, order_ref)
            container = barbican_client.containers.get(order.container_ref)

            private_key_uuid = self._retrieve_secret_uuid(
                container.secret_refs['private_key']
            )
            public_key_uuid = self._retrieve_secret_uuid(
                container.secret_refs['public_key']
            )
            self._delete_container(barbican_client, order.container_ref)
            self._delete_order(barbican_client, order_ref)
            return private_key_uuid, public_key_uuid
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error creating key pair: %s", e)
            raise exception.KeyManagerError(reason=e)

    def _get_barbican_object(
        self,
        barbican_client: barbican_client_import.Client,
        managed_obj: managed_object.ManagedObject,
    ) -> Any:
        """Converts the Castellan managed_object to a Barbican secret."""
        name = getattr(managed_obj, 'name', None)
        algorithm = getattr(managed_obj, 'algorithm', None)
        bit_length = getattr(managed_obj, 'bit_length', None)

        secret_type = self._secret_type_dict.get(type(managed_obj), 'opaque')
        encoded = managed_obj.get_encoded()
        payload = self._get_normalized_payload(encoded, secret_type)
        secret = barbican_client.secrets.create(
            payload=payload,
            algorithm=algorithm,
            bit_length=bit_length,
            name=name,
            secret_type=secret_type,
        )
        return secret

    def _get_normalized_payload(
        self, encoded_bytes: bytes | None, secret_type: str
    ) -> bytes | None:
        """Normalizes the bytes of the object.

        Barbican expects certificates, public keys, and private keys in PEM
        format, but Castellan expects these objects to be DER encoded bytes
        instead.
        """
        if encoded_bytes is None:
            return None

        if secret_type == 'public':
            pub_key = serialization.load_der_public_key(
                encoded_bytes, backend=backends.default_backend()
            )
            return pub_key.public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        elif secret_type == 'private':
            priv_key = serialization.load_der_private_key(
                encoded_bytes,
                backend=backends.default_backend(),
                password=None,
            )
            return priv_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        elif secret_type == 'certificate':
            cert = cryptography_x509.load_der_x509_certificate(
                encoded_bytes, backend=backends.default_backend()
            )
            return cert.public_bytes(encoding=serialization.Encoding.PEM)
        else:
            return encoded_bytes

    def store(
        self,
        context: Context | None,
        managed_obj: managed_object.ManagedObject,
        expiration: str | None = None,
    ) -> str:
        """Stores (i.e., registers) an object with the key manager.

        :param context: contains information of the user and the environment
            for the request
        :param managed_obj: a secret object with unencrypted payload.
            Known as "secret" to the barbicanclient api
        :param expiration: the expiration time of the secret in ISO 8601
            format
        :returns: the UUID of the stored object
        :raises KeyManagerError: if object store fails
        """
        barbican_client = self._get_barbican_client(context)

        try:
            secret = self._get_barbican_object(barbican_client, managed_obj)
            secret.expiration = expiration
            secret_ref = secret.store()
            return self._retrieve_secret_uuid(secret_ref)
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error storing object: %s", e)
            raise exception.KeyManagerError(reason=e)

    def _get_active_order(
        self, barbican_client: barbican_client_import.Client, order_ref: str
    ) -> Any:
        """Returns the order when it is active.

        Barbican key creation is done asynchronously, so this loop continues
        checking until the order is active or a timeout occurs.
        """
        active_status = 'ACTIVE'
        error_status = 'ERROR'
        number_of_retries: int = self.conf.barbican.number_of_retries
        retry_delay: int = self.conf.barbican.retry_delay
        order = barbican_client.orders.get(order_ref)
        time.sleep(0.25)
        for n in range(number_of_retries):
            if order.status == error_status:
                kwargs = {
                    "status": error_status,
                    "code": order.error_status_code,
                    "reason": order.error_reason,
                }
                msg = (
                    _(
                        "Order is in %(status)s status - status code: "
                        "%(code)s, status reason: %(reason)s"
                    )
                    % kwargs
                )
                LOG.error(msg)
                raise exception.KeyManagerError(reason=msg)
            if order.status != active_status:
                kwargs = {
                    'attempt': n,
                    'total': number_of_retries,
                    'status': order.status,
                    'active': active_status,
                    'delay': retry_delay,
                }
                msg = _(
                    "Retry attempt #%(attempt)i out of %(total)i: "
                    "Order status is '%(status)s'. Waiting for "
                    "'%(active)s', will retry in %(delay)s "
                    "seconds"
                )
                LOG.info(msg, kwargs)
                time.sleep(retry_delay)
                order = barbican_client.orders.get(order_ref)
            else:
                return order
        msg = _(
            "Exceeded retries: Failed to find '%(active)s' status "
            "within %(num_retries)i retries"
        ) % {'active': active_status, 'num_retries': number_of_retries}
        LOG.error(msg)
        raise exception.KeyManagerError(reason=msg)

    def _retrieve_secret_uuid(self, secret_ref: str) -> str:
        """Retrieves the UUID of the secret from the secret_ref.

        :param secret_ref: the href of the secret
        :return: the UUID of the secret
        """

        # The secret_ref is assumed to be of a form similar to
        # http://host:9311/v1/secrets/d152fa13-2b41-42ca-a934-6c21566c0f40
        # with the UUID at the end. This command retrieves everything
        # after the last '/', which is the UUID.
        return secret_ref.rpartition('/')[2]

    def _get_secret_data(self, secret: Any) -> bytes:
        """Retrieves the secret data.

        Converts the Barbican secret to bytes suitable for a Castellan object.
        If the secret is a public key, private key, or certificate, the secret
        is expected to be in PEM format and will be converted to DER.

        :param secret: the secret from barbican with the payload of data
        :returns: the secret data
        """
        if secret.secret_type == 'public':
            pub_key = serialization.load_pem_public_key(
                secret.payload, backend=backends.default_backend()
            )
            return pub_key.public_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        elif secret.secret_type == 'private':
            priv_key = serialization.load_pem_private_key(
                secret.payload,
                backend=backends.default_backend(),
                password=None,
            )
            return priv_key.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        elif secret.secret_type == 'certificate':
            cert = cryptography_x509.load_pem_x509_certificate(
                secret.payload, backend=backends.default_backend()
            )
            return cert.public_bytes(encoding=serialization.Encoding.DER)
        else:
            return cast(bytes, secret.payload)

    def _get_castellan_object(
        self, secret: Any, metadata_only: bool = False
    ) -> managed_object.ManagedObject:
        """Creates a Castellan managed object given the Barbican secret.

        The python barbicanclient lazy-loads the secret data, i.e. the secret
        data is not requested until secret.payload is called. If the user
        specifies metadata_only=True, secret.payload is never called,
        preventing unnecessary loading of secret data.

        :param secret: the barbican secret object
        :metadata_only: boolean indicating if the secret bytes should be
            included in the managed object
        :returns: the castellan object
        """
        secret_type: type[managed_object.ManagedObject] = op_data.OpaqueData
        for castellan_type, barbican_type in self._secret_type_dict.items():
            if barbican_type == secret.secret_type:
                secret_type = castellan_type

        secret_data: bytes | None
        if metadata_only:
            secret_data = None
        else:
            secret_data = self._get_secret_data(secret)

        object_id: str | None
        if secret.secret_ref:
            object_id = self._retrieve_secret_uuid(secret.secret_ref)
        else:
            object_id = None

        # convert created ISO8601 in Barbican to POSIX
        created: int | None
        if secret.created:
            time_stamp = timeutils.parse_isotime(
                str(secret.created)
            ).timetuple()
            created = calendar.timegm(time_stamp)
        else:
            created = None

        if issubclass(secret_type, key_base_class.Key):
            return secret_type(
                algorithm=secret.algorithm,
                bit_length=secret.bit_length,
                key=secret_data,
                name=secret.name,
                created=created,
                id=object_id,
                consumers=secret.consumers,
            )
        else:
            # Opaque Data or Passphrase
            return secret_type(  # type: ignore[misc]
                secret_data,  # type: ignore[arg-type]
                name=secret.name,
                created=created,
                id=object_id,
                consumers=secret.consumers,
            )

    def _get_secret(self, context: Context | None, object_id: str) -> Any:
        """Returns the metadata of the secret.

        :param context: contains information of the user and the environment
            for the request
        :param object_id: UUID of the secret
        :return: the secret's metadata
        :raises HTTPAuthError: if object retrieval fails with 401
        :raises HTTPClientError: if object retrieval fails with 4xx
        :raises HTTPServerError: if object retrieval fails with 5xx
        """
        if not object_id:
            raise exception.KeyManagerError('key identifier not provided')

        barbican_client = self._get_barbican_client(context)

        try:
            return barbican_client.secrets.get(object_id)
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            with excutils.save_and_reraise_exception():
                LOG.error("Error getting secret metadata: %s", e)

    def _is_secret_not_found_error(self, error: Exception) -> bool:
        if (
            isinstance(error, barbican_exceptions.HTTPClientError)
            and error.status_code == 404
        ):
            return True
        else:
            return False

    def get(
        self,
        context: Context | None,
        managed_object_id: str,
        metadata_only: bool = False,
    ) -> managed_object.ManagedObject:
        """Retrieves the specified managed object.

        :param context: contains information of the user and the environment
            for the request
        :param managed_object_id: the UUID of the object to retrieve
        :param metadata_only: whether secret data should be included
        :return: ManagedObject representation of the managed object
        :raises KeyManagerError: if object retrieval fails
        :raises ManagedObjectNotFoundError: if object not found
        """
        try:
            secret = self._get_secret(context, managed_object_id)
            return self._get_castellan_object(secret, metadata_only)
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error retrieving object: %s", e)
            if self._is_secret_not_found_error(e):
                raise exception.ManagedObjectNotFoundError(
                    uuid=managed_object_id
                )
            else:
                raise exception.KeyManagerError(reason=e)

    def delete(
        self,
        context: Context | None,
        managed_object_id: str,
        force: bool = False,
    ) -> None:
        """Deletes the specified managed object.

        :param context: contains information of the user and the environment
            for the request
        :param managed_object_id: the UUID of the object to delete
        :param force: specifies if the secret must be deleted even when they
            have consumers.
        :raises KeyManagerError: if object deletion fails
        :raises ManagedObjectNotFoundError: if the object could not be found
        """
        if not managed_object_id:
            raise exception.KeyManagerError('key identifier not provided')

        barbican_client = self._get_barbican_client(context)
        try:
            barbican_client.secrets.delete(managed_object_id, force)
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error deleting object: %s", e)
            if self._is_secret_not_found_error(e):
                raise exception.ManagedObjectNotFoundError(
                    uuid=managed_object_id
                )
            else:
                raise exception.KeyManagerError(reason=e)

    def add_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        """Add a consumer to the specified managed object

        :param context: contains information of the user and the environment
            for the request
        :param managed_object_id: the UUID of the object to update
        :param consumer_data: dict containing consumer data
        :raises KeyManagerError: if object deletion fails
        :raises ManagedObjectNotFoundError: if the object could not be found
        """

        if not managed_object_id:
            raise exception.KeyManagerError('key identifier not provided')

        barbican_client = self._get_barbican_client(context)
        try:
            barbican_client.secrets.register_consumer(
                managed_object_id, **consumer_data
            )

        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error adding consumer: %s", e)
            if self._is_secret_not_found_error(e):
                raise exception.ManagedObjectNotFoundError(
                    uuid=managed_object_id
                )
            else:
                raise exception.KeyManagerError(reason=e)

    def remove_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        barbican_client = self._get_barbican_client(context)
        try:
            barbican_client.secrets.remove_consumer(
                managed_object_id, **consumer_data
            )
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error removing consumer: %s", e)
            if self._is_secret_not_found_error(e):
                raise exception.ManagedObjectNotFoundError(
                    uuid=managed_object_id
                )
            else:
                raise exception.KeyManagerError(reason=e)

    def list(
        self,
        context: Context | None,
        object_type: type[managed_object.ManagedObject] | None = None,
        metadata_only: bool = False,
    ) -> list[managed_object.ManagedObject]:
        """Retrieves a list of managed objects that match the criteria.

        If no search criteria is given, all objects are returned.

        :param context: contains information of the user and the environment
            for the request
        :param object_type: the type of object to retrieve
        :param metadata_only: whether secret data should be included
        :raises KeyManagerError: if listing secrets fails
        """
        objects: list[managed_object.ManagedObject] = []
        barbican_client = self._get_barbican_client(context)

        if object_type and object_type not in self._secret_type_dict:
            msg = _("Invalid secret type: %s") % object_type
            LOG.error(msg)
            raise exception.KeyManagerError(reason=msg)

        secret_type = self._secret_type_dict.get(object_type)  # type: ignore[arg-type]

        try:
            secrets = barbican_client.secrets.list(secret_type=secret_type)
        except (
            barbican_exceptions.BarbicanException,
            barbican_exceptions.HTTPError,
        ) as e:
            LOG.error("Error listing objects: %s", e)
            raise exception.KeyManagerError(reason=e)

        for secret in secrets:
            try:
                obj = self._get_castellan_object(secret, metadata_only)
                objects.append(obj)
            except (
                barbican_exceptions.BarbicanException,
                barbican_exceptions.HTTPError,
            ) as e:
                LOG.warning(
                    "Error occurred while retrieving object "
                    "metadata, not adding it to the list: %s",
                    e,
                )

        return objects

    def list_options_for_discovery(
        self,
    ) -> builtins.list[tuple[str | None, builtins.list[cfg.Opt]]]:
        barbican_opts: list[cfg.Opt] = list(_barbican_opts)
        barbican_opts += loading.get_session_conf_options()
        barbican_service_user_opts = list(loading.get_session_conf_options())
        barbican_service_user_opts += loading.get_auth_common_conf_options()

        return [
            (_BARBICAN_OPT_GROUP, barbican_opts),
            (_BARBICAN_SERVICE_USER_OPT_GROUP, barbican_service_user_opts),
        ]



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the opaque data class.
"""

from castellan.common import objects
from castellan.common.objects import opaque_data
from castellan.tests import base


class OpaqueDataTestCase(base.TestCase):
    def _create_data(self):
        return opaque_data.OpaqueData(
            self.data, self.name, self.created, consumers=self.consumers
        )

    def setUp(self):
        self.data = b"secret opaque data"
        self.name = 'my opaque'
        self.created = 1448088699
        self.consumers = [
            {
                'service': 'service_test',
                'resource_type': 'type_test',
                'resource_id': 'id_test',
            }
        ]
        self.opaque_data = self._create_data()

        super().setUp()

    def test_is_not_only_metadata(self):
        self.assertFalse(self.opaque_data.is_metadata_only())

    def test_is_only_metadata(self):
        d = opaque_data.OpaqueData(None, self.name, self.created)
        self.assertTrue(d.is_metadata_only())

    def test_get_format(self):
        self.assertEqual('Opaque', self.opaque_data.format)

    def test_get_encoded(self):
        self.assertEqual(self.data, self.opaque_data.get_encoded())

    def test_get_name(self):
        self.assertEqual(self.name, self.opaque_data.name)

    def test_get_created(self):
        self.assertEqual(self.created, self.opaque_data.created)

    def test_get_consumers(self):
        self.assertEqual(self.consumers, self.opaque_data.consumers)

    def test_get_created_none(self):
        created = None
        data = opaque_data.OpaqueData(
            self.data, self.name, created, consumers=self.consumers
        )

        self.assertEqual(created, data.created)

    def test___eq__(self):
        self.assertTrue(self.opaque_data == self.opaque_data)
        self.assertTrue(self.opaque_data is self.opaque_data)

        self.assertFalse(self.opaque_data is None)
        self.assertFalse(None == self.opaque_data)  # noqa: E711

        other_opaque_data = opaque_data.OpaqueData(self.data)
        self.assertTrue(self.opaque_data == other_opaque_data)
        self.assertFalse(self.opaque_data is other_opaque_data)

    def test___ne___none(self):
        self.assertTrue(self.opaque_data is not None)
        self.assertTrue(None != self.opaque_data)  # noqa: E711

    def test___ne___data(self):
        other_opaque = opaque_data.OpaqueData(b'other data', self.name)
        self.assertTrue(self.opaque_data != other_opaque)

    def test_to_and_from_dict(self):
        other = objects.from_dict(self.opaque_data.to_dict())
        self.assertEqual(self.opaque_data, other)



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from typing import cast

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from stevedore import driver
from stevedore import exception

from castellan.key_manager import key_manager
from castellan.key_manager import migration

LOG = logging.getLogger(__name__)

key_manager_opts = [
    cfg.StrOpt(
        'backend',
        default='barbican',
        deprecated_name='api_class',
        deprecated_group='key_manager',
        help='Specify the key manager implementation. Options are '
        '"barbican" and "vault". Default is  "barbican". Will '
        'support the  values earlier set using '
        '[key_manager]/api_class for some time.',
    ),
]


def API(
    configuration: cfg.ConfigOpts | None = None,
) -> key_manager.KeyManager:
    conf = configuration or cfg.CONF
    conf.register_opts(key_manager_opts, group='key_manager')

    mgr: driver.DriverManager[key_manager.KeyManager]
    try:
        mgr = driver.DriverManager(
            "castellan.drivers",
            conf.key_manager.backend,
            invoke_on_load=True,
            invoke_args=(conf,),
        )
        # we know the return type will be an instance since invoke_on_load is
        # true: stevedore just needs better hints
        key_mgr = cast(key_manager.KeyManager, mgr.driver)
    except exception.NoMatches:
        LOG.warning(
            "Deprecation Warning : %s is not a stevedore based driver, "
            "trying to load it as a class",
            conf.key_manager.backend,
        )
        cls = importutils.import_class(conf.key_manager.backend)
        key_mgr = cls(configuration=conf)

    return migration.handle_migration(conf, key_mgr)



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the private key class.
"""

from castellan.common import objects
from castellan.common.objects import private_key
from castellan.tests import base
from castellan.tests import utils


class PrivateKeyTestCase(base.KeyTestCase):
    def _create_key(self):
        return private_key.PrivateKey(
            self.algorithm,
            self.bit_length,
            self.encoded,
            self.name,
            self.created,
            consumers=self.consumers,
        )

    def setUp(self):
        self.algorithm = 'RSA'
        self.bit_length = 2048
        self.encoded = bytes(utils.get_private_key_der())
        self.name = 'my key'
        self.created = 1448088699
        self.consumers = [
            {
                'service': 'service_test',
                'resource_type': 'type_test',
                'resource_id': 'id_test',
            }
        ]

        super().setUp()

    def test_is_not_only_metadata(self):
        self.assertFalse(self.key.is_metadata_only())

    def test_is_only_metadata(self):
        k = private_key.PrivateKey(
            self.algorithm, self.bit_length, None, self.name, self.created
        )
        self.assertTrue(k.is_metadata_only())

    def test_get_algorithm(self):
        self.assertEqual(self.algorithm, self.key.algorithm)

    def test_get_bit_length(self):
        self.assertEqual(self.bit_length, self.key.bit_length)

    def test_get_name(self):
        self.assertEqual(self.name, self.key.name)

    def test_get_format(self):
        self.assertEqual('PKCS8', self.key.format)

    def test_get_encoded(self):
        self.assertEqual(self.encoded, self.key.get_encoded())

    def test_get_created(self):
        self.assertEqual(self.created, self.key.created)

    def test_get_consumers(self):
        self.assertEqual(self.consumers, self.key.consumers)

    def test_get_created_none(self):
        created = None
        key = private_key.PrivateKey(
            self.algorithm,
            self.bit_length,
            self.encoded,
            self.name,
            created,
            consumers=self.consumers,
        )

        self.assertEqual(created, key.created)

    def test___eq__(self):
        self.assertTrue(self.key == self.key)
        self.assertTrue(self.key is self.key)

        self.assertFalse(self.key is None)
        self.assertFalse(None == self.key)  # noqa: E711

        other_key = private_key.PrivateKey(
            self.algorithm,
            self.bit_length,
            self.encoded,
            consumers=self.consumers,
        )
        self.assertTrue(self.key == other_key)
        self.assertFalse(self.key is other_key)

    def test___ne___none(self):
        self.assertTrue(self.key is not None)
        self.assertTrue(None != self.key)  # noqa: E711

    def test___ne___algorithm(self):
        other_key = private_key.PrivateKey(
            'DSA',
            self.bit_length,
            self.encoded,
            self.name,
            consumers=self.consumers,
        )
        self.assertTrue(self.key != other_key)

    def test___ne___bit_length(self):
        other_key = private_key.PrivateKey(
            self.algorithm,
            4096,
            self.encoded,
            self.name,
            consumers=self.consumers,
        )
        self.assertTrue(self.key != other_key)

    def test___ne___encoded(self):
        different_encoded = bytes(utils.get_private_key_der()) + b'\x00'
        other_key = private_key.PrivateKey(
            self.algorithm,
            self.bit_length,
            different_encoded,
            self.name,
            consumers=self.consumers,
        )
        self.assertTrue(self.key != other_key)

    def test___ne___consumers(self):
        different_consumers = [
            {
                'service': 'other_service',
                'resource_type': 'other_type',
                'resource_id': 'other_id',
            }
        ]
        other_key = private_key.PrivateKey(
            self.algorithm,
            self.bit_length,
            self.encoded,
            self.name,
            consumers=different_consumers,
        )

        self.assertTrue(self.key is not other_key)

    def test_to_and_from_dict(self):
        other = objects.from_dict(self.key.to_dict())
        self.assertEqual(self.key, other)



# Copyright (c) 2015 IBM
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the password credential
"""

from castellan.common.credentials import password
from castellan.tests import base


class PasswordTestCase(base.TestCase):
    def _create_password_credential(self):
        return password.Password(self.username, self.password)

    def setUp(self):
        self.username = "admin"
        self.password = "Pa$$w0rd1"

        self.password_credential = self._create_password_credential()

        super().setUp()

    def test_get_username(self):
        self.assertEqual(self.username, self.password_credential.username)

    def test_get_password(self):
        self.assertEqual(self.password, self.password_credential.password)

    def test___eq__(self):
        self.assertTrue(self.password_credential == self.password_credential)
        self.assertTrue(self.password_credential is self.password_credential)

        self.assertFalse(self.password_credential is None)
        self.assertFalse(None == self.password_credential)  # noqa: E711

        other_password_credential = password.Password(
            self.username, self.password
        )
        self.assertTrue(self.password_credential == other_password_credential)
        self.assertFalse(self.password_credential is other_password_credential)

    def test___ne___none(self):
        self.assertTrue(self.password_credential is not None)
        self.assertTrue(None != self.password_credential)  # noqa: E711

    def test___ne___username(self):
        other_username = "service"
        other_password_credential = password.Password(
            other_username, self.password
        )
        self.assertTrue(self.password_credential != other_password_credential)

    def test___ne___password(self):
        other_password = "i143Cats"
        other_password_credential = password.Password(
            self.username, other_password
        )
        self.assertTrue(self.password_credential != other_password_credential)



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Key manager API
"""

from __future__ import annotations

from typing import TypeAlias

import abc
import builtins

from oslo_config import cfg
from oslo_context import context as ctx

from castellan.common.credentials import credential
from castellan.common.objects import managed_object
from castellan.common.objects import opaque_data as op_data
from castellan.common.objects import passphrase
from castellan.common.objects import private_key as pri_key
from castellan.common.objects import public_key as pub_key
from castellan.common.objects import symmetric_key as sym_key
from castellan.common.objects import x_509

Context: TypeAlias = credential.Credential | ctx.RequestContext


class KeyManager(metaclass=abc.ABCMeta):
    """Base Key Manager Interface

    A Key Manager is responsible for managing encryption keys for volumes. A
    Key Manager is responsible for creating, reading, and deleting keys.
    """

    _secret_type_dict: dict[type[managed_object.ManagedObject], str] = {
        op_data.OpaqueData: "opaque",
        passphrase.Passphrase: "passphrase",
        pri_key.PrivateKey: "private",
        pub_key.PublicKey: "public",
        sym_key.SymmetricKey: "symmetric",
        x_509.X509: "certificate",
    }

    @abc.abstractmethod
    def __init__(self, configuration: cfg.ConfigOpts | None) -> None:
        """Instantiate a KeyManager object.

        Creates a KeyManager object with implementation specific details
        obtained from the supplied configuration.
        """
        pass

    @abc.abstractmethod
    def create_key(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> str:
        """Creates a symmetric key.

        This method creates a symmetric key and returns the key's UUID. If the
        specified context does not permit the creation of keys, then a
        NotAuthorized exception should be raised.
        """
        pass

    @abc.abstractmethod
    def create_key_pair(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> tuple[str, str]:
        """Creates an asymmetric key pair.

        This method creates an asymmetric key pair and returns the pair of key
        UUIDs. If the specified context does not permit the creation of keys,
        then a NotAuthorized exception should be raised. The order of the UUIDs
        will be (private, public).
        """
        pass

    @abc.abstractmethod
    def store(
        self,
        context: Context | None,
        managed_object: managed_object.ManagedObject,
        expiration: str | None = None,
    ) -> str:
        """Stores a managed object with the key manager.

        This method stores the specified managed object and returns its UUID
        that identifies it within the key manager. If the specified context
        does not permit the creation of keys, then a NotAuthorized exception
        should be raised.
        """
        pass

    @abc.abstractmethod
    def get(
        self,
        context: Context | None,
        managed_object_id: str,
        metadata_only: bool = False,
    ) -> managed_object.ManagedObject:
        """Retrieves the specified managed object.

        Implementations should verify that the caller has permissions to
        retrieve the managed object by checking the context object passed in
        as context. If the user lacks permission then a NotAuthorized
        exception is raised.

        If the caller requests only metadata, then the object that is
        returned will contain only the secret metadata and no secret bytes.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling
        this method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.
        """
        pass

    @abc.abstractmethod
    def delete(
        self,
        context: Context | None,
        managed_object_id: str,
        force: bool = False,
    ) -> None:
        """Deletes the specified managed object.

        Implementations should verify that the caller has permission to delete
        the managed object by checking the context object (context). A
        NotAuthorized exception should be raised if the caller lacks
        permission.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling this
        method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.

        Implementations that block the deletion of secrets with consumers
        without making the "force" parameter equals True should raise
        an exception.
        """
        pass

    @abc.abstractmethod
    def add_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        """Add a consumer to a managed object.

        Implementations should verify that the caller has permission to
        add a consumer to the managed object by checking the context object
        (context). A NotAuthorized exception should be raised if the caller
        lacks permission.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling this
        method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.
        """
        pass

    @abc.abstractmethod
    def remove_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> None:
        """Remove a consumer from a managed object.

        Implementations should verify that the caller has permission to
        remove a consumer to the managed object by checking the context object
        (context). A NotAuthorized exception should be raised if the caller
        lacks permission.

        If the specified object does not exist, then a KeyError should be
        raised. Implementations should preclude users from discerning the
        UUIDs of objects that belong to other users by repeatedly calling this
        method. That is, objects that belong to other users should be
        considered "non-existent" and completely invisible.
        """
        pass

    def list(
        self,
        context: Context | None,
        object_type: type[managed_object.ManagedObject] | None = None,
        metadata_only: bool = False,
    ) -> list[managed_object.ManagedObject]:
        """Lists the managed objects given the criteria.

        Implementations should verify that the caller has permission to list
        the managed objects and should only list the objects the caller has
        access to by checking the context object (context). A NotAuthorized
        exception should be raised if the caller lacks permission.

        A list of managed objects or managed object metadata should be
        returned, depending on the metadata_only flag. If no objects are
        found, an empty list should be returned instead.
        """
        return []

    def list_options_for_discovery(
        self,
    ) -> builtins.list[tuple[str | None, builtins.list[cfg.Opt]]]:
        """Lists the KeyManager's configure options.

        KeyManagers should advertise all supported options through this
        method for the purpose of sample generation by oslo-config-generator.
        Each item in the advertised list should be tuple composed by the group
        name and a list of options belonging to that group. None should be used
        as the group name for the DEFAULT group.

        :returns: the list of supported options of a KeyManager.
        """
        return []



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for the not implemented key manager.
"""

from castellan.key_manager import not_implemented_key_manager
from castellan.tests.unit.key_manager import test_key_manager


class NotImplementedKeyManagerTestCase(test_key_manager.KeyManagerTestCase):
    def _create_key_manager(self):
        return not_implemented_key_manager.NotImplementedKeyManager()

    def test_create_key(self):
        self.assertRaises(NotImplementedError, self.key_mgr.create_key, None)

    def test_create_key_pair(self):
        self.assertRaises(
            NotImplementedError, self.key_mgr.create_key_pair, None, None, None
        )

    def test_store(self):
        self.assertRaises(NotImplementedError, self.key_mgr.store, None, None)

    def test_copy(self):
        self.assertRaises(NotImplementedError, self.key_mgr.copy, None, None)

    def test_get(self):
        self.assertRaises(NotImplementedError, self.key_mgr.get, None, None)

    def test_list(self):
        self.assertRaises(NotImplementedError, self.key_mgr.list, None)

    def test_delete(self):
        self.assertRaises(NotImplementedError, self.key_mgr.delete, None, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Key manager implementation for Vault
"""

from __future__ import annotations

import binascii
import builtins
from collections.abc import Callable
import datetime
import os
import time
from typing import Any
from typing import NoReturn
import uuid

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import PrivateFormat
from cryptography.hazmat.primitives.serialization import PublicFormat
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils
import requests

from castellan.common import exception
from castellan.common.objects import managed_object
from castellan.common.objects import private_key as pri_key
from castellan.common.objects import public_key as pub_key
from castellan.common.objects import symmetric_key as sym_key
from castellan.i18n import _
from castellan.key_manager import key_manager
from castellan.key_manager.key_manager import Context

_DEFAULT_VAULT_URL = "http://127.0.0.1:8200"
_DEFAULT_MOUNTPOINT = "secret"
_DEFAULT_VERSION = 2

_vault_opts = [
    cfg.StrOpt('root_token_id', secret=True, help='root token for vault'),
    cfg.StrOpt(
        'auth_method',
        default='approle',
        choices=('approle', 'jwt', 'kubernetes'),
        help='Auth method to use when connecting to Vault. '
        '"approle" uses approle_role_id and approle_secret_id. '
        '"jwt" and "kubernetes" use token_role and token_file.',
    ),
    cfg.StrOpt(
        'approle_role_id',
        secret=True,
        help='AppRole role_id for authentication with vault',
    ),
    cfg.StrOpt(
        'approle_secret_id',
        secret=True,
        help='AppRole secret_id for authentication with vault',
    ),
    cfg.StrOpt(
        'token_role',
        help='Vault role name for token-based auth. '
        'Required when auth_method is jwt or kubernetes.',
    ),
    cfg.StrOpt(
        'token_file',
        help='Path to the token file used for Vault login. '
        'Required when auth_method is jwt or kubernetes.',
    ),
    cfg.StrOpt(
        'auth_path',
        help='Mount path of the Vault auth backend, used in '
        'the login URL /v1/auth/<auth_path>/login. '
        'Defaults to the value of auth_method when not set. '
        'Override this when the auth backend is mounted at '
        'a non-default path '
        '(e.g. "kubernetes-my-cluster" instead of '
        '"kubernetes").',
    ),
    cfg.StrOpt(
        'kv_mountpoint',
        default=_DEFAULT_MOUNTPOINT,
        help='Mountpoint of KV store in Vault to use',
    ),
    cfg.StrOpt(
        'kv_path', help='Path relative to root of KV store in Vault to use.'
    ),
    cfg.IntOpt(
        'kv_version',
        default=_DEFAULT_VERSION,
        choices=(1, 2),
        help='Version of KV store in Vault to use.',
    ),
    cfg.URIOpt(
        'vault_url',
        default=_DEFAULT_VAULT_URL,
        schemes=['http', 'https'],
        help='Use this endpoint to connect to Vault',
    ),
    cfg.StrOpt('ssl_ca_crt_file', help='Absolute path to ca cert file'),
    cfg.BoolOpt(
        'use_ssl',
        default=False,
        deprecated_for_removal=True,
        deprecated_reason='This option has no effect.',
        help=_('SSL Enabled/Disabled'),
    ),
    cfg.StrOpt(
        "namespace",
        help=_(
            "Vault Namespace to use for all requests to Vault. "
            "Vault Namespaces feature is available only in "
            "Vault Enterprise"
        ),
    ),
    cfg.FloatOpt(
        'timeout',
        default=60,
        help=_('Timeout (in seconds) in each request to Vault'),
    ),
]

_VAULT_OPT_GROUP = 'vault'

_EXCEPTIONS_BY_CODE = [
    requests.codes['internal_server_error'],
    requests.codes['service_unavailable'],
    requests.codes['request_timeout'],
    requests.codes['gateway_timeout'],
    requests.codes['precondition_failed'],
]

LOG = logging.getLogger(__name__)


class VaultKeyManager(key_manager.KeyManager):
    """Key Manager Interface that wraps the Vault REST API."""

    _conf: cfg.ConfigOpts
    _session: requests.Session
    _root_token_id: str | None
    _approle_role_id: str | None
    _approle_secret_id: str | None
    _auth_method: str
    _token_role: str | None
    _token_file: str
    _auth_path: str
    _cached_token_id: str | None
    _cached_token_ttl: int | None
    _cached_token_issue: datetime.datetime | None
    _kv_mountpoint: str
    _kv_path: str | None
    _kv_version: int
    _vault_url: str
    _namespace: str | None
    _timeout: float
    _verify_server: str | bool

    def __init__(self, configuration: cfg.ConfigOpts) -> None:
        self._conf = configuration
        self._conf.register_opts(_vault_opts, group=_VAULT_OPT_GROUP)
        self._session = requests.Session()
        self._root_token_id = self._conf.vault.root_token_id
        self._approle_role_id = self._conf.vault.approle_role_id
        self._approle_secret_id = self._conf.vault.approle_secret_id
        self._auth_method = self._conf.vault.auth_method
        self._token_role = self._conf.vault.token_role
        self._token_file = self._conf.vault.token_file
        self._auth_path = self._conf.vault.auth_path or self._auth_method
        self._cached_token_id = None
        self._cached_token_ttl = None
        self._cached_token_issue = None
        self._kv_mountpoint = self._conf.vault.kv_mountpoint
        self._kv_path = self._conf.vault.kv_path
        self._kv_version = self._conf.vault.kv_version
        self._vault_url = self._conf.vault.vault_url
        self._namespace = self._conf.vault.namespace
        self._timeout = self._conf.vault.timeout
        if self._vault_url.startswith("https://"):
            self._verify_server = self._conf.vault.ssl_ca_crt_file or True
        else:
            self._verify_server = False

    def _get_url(self) -> str:
        if not self._vault_url.endswith('/'):
            self._vault_url += '/'
        return self._vault_url

    def _get_resource_url(self, key_id: str | None = None) -> str:
        return '{}v1/{}/{}{}{}'.format(
            self._get_url(),
            self._kv_mountpoint,
            # no key_id is for listing and 'data/' doesn't works
            ''
            if self._kv_version == 1
            else 'data/'
            if key_id
            else 'metadata/',
            (self._kv_path + '/') if self._kv_path else '',
            key_id if key_id else '?list=true',
        )

    @property
    def _cached_token(self) -> str | None:
        if (
            self._cached_token_issue is not None
            and self._cached_token_ttl is not None
            and timeutils.is_older_than(
                self._cached_token_issue, self._cached_token_ttl
            )
        ):
            self._cached_token_id = None
        return self._cached_token_id

    def _set_namespace(self, headers: dict[str, str]) -> dict[str, str]:
        if self._namespace:
            headers["X-Vault-Namespace"] = self._namespace
        return headers

    def _build_auth_headers(self) -> dict[str, str]:
        if self._root_token_id:
            return self._set_namespace({'X-Vault-Token': self._root_token_id})

        if self._cached_token:
            return self._set_namespace({'X-Vault-Token': self._cached_token})

        login_url = f'{self._get_url()}v1/auth/{self._auth_path}/login'

        if self._auth_method == 'approle':
            if not self._approle_role_id:
                return {}
            params: dict[str, str] = {'role_id': self._approle_role_id}
            if self._approle_secret_id:
                params['secret_id'] = self._approle_secret_id
            return self._vault_login(login_url, params)

        elif self._auth_method in ('jwt', 'kubernetes'):
            if not self._token_role:
                raise exception.KeyManagerError(
                    "token_role is required when "
                    f"auth_method={self._auth_method}"
                )
            if not self._token_file:
                raise exception.KeyManagerError(
                    "token_file is required when "
                    f"auth_method={self._auth_method}"
                )
            try:
                with open(self._token_file) as f:
                    token = f.read().strip()
            except OSError as ex:
                raise exception.KeyManagerError(
                    f"Failed to read token file from {self._token_file}: {ex}"
                )

            params = {
                'role': self._token_role,
                'jwt': token,
            }
            return self._vault_login(login_url, params)

        raise exception.KeyManagerError(
            f"Unknown auth_method: {self._auth_method}"
        )

    def _vault_login(
        self,
        login_url: str,
        params: dict[str, str],
    ) -> dict[str, str]:
        """Authenticate with Vault and cache the resulting token.

        Shared by AppRole and token-based (Kubernetes, JWT/OIDC) auth methods.
        """
        token_issue_utc = timeutils.utcnow()
        headers = self._set_namespace({})
        try:
            resp = self._session.post(
                url=login_url,
                json=params,
                headers=headers,
                verify=self._verify_server,
                timeout=self._timeout,
            )
        except Exception as ex:
            raise exception.KeyManagerError(str(ex))

        if resp.status_code in _EXCEPTIONS_BY_CODE:
            raise exception.KeyManagerError(resp.reason)
        if resp.status_code == requests.codes['forbidden']:
            raise exception.Forbidden()

        resp_data = resp.json()

        if resp.status_code == requests.codes['bad_request']:
            raise exception.KeyManagerError(', '.join(resp_data['errors']))

        self._cached_token_id = resp_data['auth']['client_token']
        self._cached_token_issue = token_issue_utc
        self._cached_token_ttl = resp_data['auth']['lease_duration']
        return self._set_namespace({'X-Vault-Token': self._cached_token_id})

    def _do_http_request(
        self,
        method: Callable[..., requests.Response],
        resource: str,
        json: dict[str, Any] | None = None,
    ) -> requests.Response:
        headers = self._build_auth_headers()

        try:
            resp = method(
                resource,
                headers=headers,
                json=json,
                verify=self._verify_server,
                timeout=self._timeout,
            )
        except Exception as ex:
            raise exception.KeyManagerError(str(ex))

        if resp.status_code in _EXCEPTIONS_BY_CODE:
            raise exception.KeyManagerError(resp.reason)
        if resp.status_code == requests.codes['forbidden']:
            raise exception.Forbidden()

        return resp

    def create_key_pair(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> tuple[str, str]:
        """Creates an asymmetric key pair."""

        if algorithm.lower() != 'rsa':
            raise NotImplementedError(
                "VaultKeyManager only implements rsa keys"
            )

        priv_key = rsa.generate_private_key(
            public_exponent=65537, key_size=length, backend=default_backend()
        )

        private_key = pri_key.PrivateKey(
            'RSA',
            length,
            priv_key.private_bytes(
                Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()
            ),
        )

        private_key_id = uuid.uuid4().hex
        private_id = self._store_key_value(private_key_id, private_key)

        # pub_key = priv_key.public_key()
        public_key = pub_key.PublicKey(
            'RSA',
            length,
            priv_key.public_key().public_bytes(
                Encoding.PEM, PublicFormat.SubjectPublicKeyInfo
            ),
        )

        public_key_id = uuid.uuid4().hex
        public_id = self._store_key_value(public_key_id, public_key)

        return private_id, public_id

    def _store_key_value(
        self, key_id: str, value: managed_object.ManagedObject
    ) -> str:

        type_value = self._secret_type_dict.get(type(value))
        if type_value is None:
            raise exception.KeyManagerError(
                f"Unknown type for value : {value!r}"
            )

        encoded = value.get_encoded()
        if encoded is None:
            raise exception.KeyManagerError("Cannot store object without data")

        record: dict[str, Any] = {
            'type': type_value,
            'value': binascii.hexlify(encoded).decode('utf-8'),
            'algorithm': (
                value.algorithm if hasattr(value, 'algorithm') else None
            ),
            'bit_length': (
                value.bit_length if hasattr(value, 'bit_length') else None
            ),
            'name': value.name,
            'created': value.created,
        }
        if self._kv_version > 1:
            record = {'data': record}

        self._do_http_request(
            self._session.post, self._get_resource_url(key_id), json=record
        )

        return key_id

    def create_key(
        self,
        context: Context | None,
        algorithm: str,
        length: int,
        expiration: str | None = None,
        name: str | None = None,
    ) -> str:
        """Creates a symmetric key."""

        if length % 8:
            msg = _("Length must be multiple of 8.")
            raise ValueError(msg)

        key_id = uuid.uuid4().hex
        key_value = os.urandom((length or 256) // 8)
        key = sym_key.SymmetricKey(
            algorithm,
            length or 256,
            key_value,
            key_id,
            # FIXME(stephenfin): This should be an int yet we pass name (a str)
            # if set?
            name or int(time.time()),  # type: ignore[arg-type]
        )

        return self._store_key_value(key_id, key)

    def store(
        self,
        context: Context | None,
        managed_obj: managed_object.ManagedObject,
        expiration: str | None = None,
    ) -> str:
        """Stores (i.e., registers) an object with the key manager.

        :param context: contains information of the user and the environment
            for the request
        :param managed_obj: a secret object with unencrypted payload.
        :param expiration: (currently ignored)
        :returns: the UUID of the stored object
        :raises KeyManagerError: if object store fails
        """
        key_id = uuid.uuid4().hex
        return self._store_key_value(key_id, managed_obj)

    def get(
        self,
        context: Context | None,
        managed_object_id: str,
        metadata_only: bool = False,
    ) -> managed_object.ManagedObject:
        """Retrieves the key identified by the specified id."""

        if not managed_object_id:
            raise exception.KeyManagerError('key identifier not provided')

        resp = self._do_http_request(
            self._session.get, self._get_resource_url(managed_object_id)
        )

        if resp.status_code == requests.codes['not_found']:
            raise exception.ManagedObjectNotFoundError(uuid=managed_object_id)

        record = resp.json()['data']
        if self._kv_version > 1:
            record = record['data']

        key = None if metadata_only else binascii.unhexlify(record['value'])

        clazz: type[managed_object.ManagedObject] | None = None
        for type_clazz, type_name in self._secret_type_dict.items():
            if type_name == record['type']:
                clazz = type_clazz

        if clazz is None:
            raise exception.KeyManagerError(
                "Unknown type : {!r}".format(record['type'])
            )

        # TODO(stephenfin): Use isinstance checks instead of hasattr checks
        if hasattr(clazz, 'algorithm') and hasattr(clazz, 'bit_length'):
            return clazz(  # type: ignore[call-arg]
                record['algorithm'],
                record['bit_length'],
                key,  # type: ignore[arg-type]
                record['name'],
                record['created'],
                managed_object_id,
            )
        else:
            return clazz(
                key,  # type: ignore[arg-type]
                record['name'],
                record['created'],
                managed_object_id,  # type: ignore[arg-type]
            )

    def delete(
        self,
        context: Context | None,
        managed_object_id: str,
        force: bool = False,
    ) -> None:
        """Represents deleting the key.

        The 'force' parameter is not used whatsoever and only kept to allow
        consistency with the Barbican implementation.
        """

        if not managed_object_id:
            raise exception.KeyManagerError('key identifier not provided')

        resp = self._do_http_request(
            self._session.delete, self._get_resource_url(managed_object_id)
        )

        if resp.status_code == requests.codes['not_found']:
            raise exception.ManagedObjectNotFoundError(uuid=managed_object_id)

    def add_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> NoReturn:
        raise NotImplementedError(
            "VaultKeyManager does not implement adding consumers"
        )

    def remove_consumer(
        self,
        context: Context | None,
        managed_object_id: str,
        consumer_data: dict[str, str],
    ) -> NoReturn:
        raise NotImplementedError(
            "VaultKeyManager does not implement deleting consumers"
        )

    def list(
        self,
        context: Context | None,
        object_type: type[managed_object.ManagedObject] | None = None,
        metadata_only: bool = False,
    ) -> list[managed_object.ManagedObject]:
        """Lists the managed objects given the criteria."""

        if object_type and object_type not in self._secret_type_dict:
            msg = _("Invalid secret type: %s") % object_type
            raise exception.KeyManagerError(reason=msg)

        resp = self._do_http_request(
            self._session.get, self._get_resource_url()
        )

        keys: list[str]
        if resp.status_code == requests.codes['not_found']:
            keys = []
        else:
            keys = resp.json()['data']['keys']

        objects: list[managed_object.ManagedObject] = []
        for obj_id in keys:
            try:
                obj = self.get(context, obj_id, metadata_only=metadata_only)
                if object_type is None or isinstance(obj, object_type):
                    objects.append(obj)
            except exception.ManagedObjectNotFoundError as e:
                LOG.warning(
                    "Error occurred while retrieving object "
                    "metadata, not adding it to the list: %s",
                    e,
                )
                pass
        return objects

    def list_options_for_discovery(
        self,
    ) -> builtins.list[tuple[str | None, builtins.list[cfg.Opt]]]:
        return [(_VAULT_OPT_GROUP, _vault_opts)]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

r"""
Castellan Oslo Config Driver
----------------------------

This driver is an oslo.config backend driver implemented with Castellan. It
extends oslo.config's capabilities by enabling it to retrieve configuration
values from a secret manager behind Castellan.

The setup of a Castellan configuration source is as follow::

    [DEFAULT]
    config_source = castellan_config_group

    [castellan_config_group]
    driver = castellan
    config_file = castellan.conf
    mapping_file = mapping.conf

In the following sessions, you can find more information about this driver's
classes and its options.

The Driver Class
================

.. autoclass:: CastellanConfigurationSourceDriver

The Configuration Source Class
==============================

.. autoclass:: CastellanConfigurationSource

"""

from typing import Any

from oslo_config import cfg
from oslo_config import sources
from oslo_log import log

from castellan.common.credentials import credential
from castellan.common.exception import AuthTypeInvalidError
from castellan.common.exception import KeyManagerError
from castellan.common.exception import ManagedObjectNotFoundError
from castellan.common import utils
from castellan import key_manager
from castellan.key_manager import key_manager as km

LOG = log.getLogger(__name__)


class CastellanConfigurationSourceDriver(sources.ConfigurationSourceDriver):
    """A backend driver for configuration values served through castellan.

    Required options:
      - config_file: The castellan configuration file.

      - mapping_file: A configuration/castellan_id mapping file. This file
                      creates connections between configuration options and
                      castellan ids. The group and option name remains the
                      same, while the value gets stored a secret manager behind
                      castellan and is replaced by its castellan id. The ids
                      will be used to fetch the values through castellan.
    """

    _castellan_driver_opts: list[cfg.Opt] = [
        cfg.StrOpt(
            'config_file',
            required=True,
            sample_default='etc/castellan/castellan.conf',
            help=('The path to a castellan configuration file.'),
        ),
        cfg.StrOpt(
            'mapping_file',
            required=True,
            sample_default='etc/castellan/secrets_mapping.conf',
            help=('The path to a configuration/castellan_id mapping file.'),
        ),
    ]

    def list_options_for_discovery(self) -> list[cfg.Opt]:
        return self._castellan_driver_opts

    def open_source_from_opt_group(
        self, conf: cfg.ConfigOpts, group_name: str
    ) -> "CastellanConfigurationSource":
        conf.register_opts(self._castellan_driver_opts, group_name)

        return CastellanConfigurationSource(
            group_name,
            conf[group_name].config_file,
            conf[group_name].mapping_file,
        )


class CastellanConfigurationSource(sources.ConfigurationSource):
    """A configuration source for values served through castellan.

    :param config_file: The path to a castellan configuration file.
    :param mapping_file: The path to a configuration/castellan_id mapping file.
    """

    _mngr: km.KeyManager
    _mapping: dict[str, dict[str, list[str]]]
    _context: credential.Credential | None

    def __init__(
        self, group_name: str, config_file: str, mapping_file: str
    ) -> None:
        conf = cfg.ConfigOpts()
        conf(args=[], default_config_files=[config_file])

        self._name = group_name
        self._mngr = key_manager.API(conf)
        self._mapping = {}

        try:
            self._context = utils.credential_factory(conf)
        except AuthTypeInvalidError:
            self._context = None

            LOG.warning(
                "Invalid 'auth_type' in '%s', auth_type: %s. "
                "Context set to 'None'. Supported 'auth_type' values: "
                "'token', 'password', 'keystone_token', "
                "'keystone_password'.",
                config_file,
                conf.key_manager.auth_type,
            )

        cfg.ConfigParser(mapping_file, self._mapping).parse()

    def get(
        self, group_name: str | None, option_name: str, opt: cfg.Opt
    ) -> tuple[Any, cfg.LocationInfo | None]:
        try:
            group_name = group_name or "DEFAULT"

            castellan_id = self._mapping[group_name][option_name][0]

            encoded = self._mngr.get(self._context, castellan_id).get_encoded()
            if encoded is None:
                return (sources._NoValue, None)

            return (
                encoded.decode(),
                cfg.LocationInfo(cfg.Locations.user, castellan_id),
            )

        except KeyError:
            # no mapping 'option = castellan_id'
            LOG.debug(
                "option '[%s] %s' not present in '[%s] mapping_file'",
                group_name,
                option_name,
                self._name,
            )

        except KeyManagerError:
            # error retrieving the secret from the key manager
            LOG.exception(
                "Failed to retrieve secret for option '[%s] %s' in "
                "'[%s] mapping_file'",
                group_name,
                option_name,
                self._name,
            )

        except ManagedObjectNotFoundError:
            # good mapping, but unknown castellan_id by secret manager
            LOG.error(
                "invalid castellan_id for option '[%s] %s' in '[%s] "
                "mapping_file'",
                group_name,
                option_name,
                self._name,
            )

        return (sources._NoValue, None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import cast
import warnings

import pbr.version


def __getattr__(name: str) -> str:
    if name == '__version__':
        warnings.warn(
            "Accessing castellan.__version__ is deprecated and will be "
            "removed in a future release. Use importlib.metadata instead: "
            "importlib.metadata.version('castellan')",
            DeprecationWarning,
            stacklevel=2,
        )
        return cast(str, pbr.version.VersionInfo('castellan').version_string())
    raise AttributeError(f"module 'castellan' has no attribute {name!r}")



# Copyright (c) 2015 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base Key Class

This module defines the Key class. The Key class is the base class to
represent all encryption keys. The basis for this class was copied
from Java.
"""

from __future__ import annotations

import abc
import binascii
from typing import Any

from castellan.common import exception
from castellan.common.objects import managed_object


class Key(managed_object.ManagedObject):
    """Base class to represent all keys."""

    def __init__(
        self,
        algorithm: str,
        bit_length: int,
        key: bytes | None,
        name: str | None = None,
        created: int | None = None,
        id: str | None = None,
        consumers: list[dict[str, str]] | None = None,
    ) -> None:
        self._alg = algorithm
        self._bit_length = bit_length
        self._key = key
        super().__init__(
            name=name, created=created, id=id, consumers=consumers
        )

    @property
    @abc.abstractmethod
    def algorithm(self) -> str:
        """Returns the key's algorithm.

        Returns the key's algorithm. For example, "DSA" indicates that this key
        is a DSA key and "AES" indicates that this key is an AES key.
        """
        pass

    @property
    @abc.abstractmethod
    def bit_length(self) -> int:
        """Returns the key's bit length.

        Returns the key's bit length. For example, for AES symmetric keys,
        this refers to the length of the key, and for RSA keys, this refers to
        the length of the modulus.
        """
        pass

    def to_dict(self, metadata_only: bool = False) -> dict[str, Any]:
        dict_fields = super().to_dict(metadata_only)

        dict_fields["algorithm"] = self.algorithm
        dict_fields["bit_length"] = self.bit_length
        dict_fields["consumers"] = self.consumers

        return dict_fields

    @classmethod
    def from_dict(
        cls,
        dict_fields: dict[str, Any],
        id: str | None = None,
        metadata_only: bool = False,
        consumers: list[dict[str, str]] | None = None,
    ) -> Key:
        try:
            value = None

            # NOTE(moguimar): the managed object's value is exported as
            # a hex string. For now, this is a compatibility thing with
            # the already existent vault_key_manager backend.
            if not metadata_only and dict_fields["value"] is not None:
                value = binascii.unhexlify(dict_fields["value"])

            # NOTE: The from_dict is designed to be called on subclasses
            # which have different __init__ signatures
            return cls(
                algorithm=dict_fields["algorithm"],
                bit_length=dict_fields["bit_length"],
                key=value,
                name=dict_fields["name"],
                created=dict_fields["created"],
                id=id,
                consumers=dict_fields["consumers"] or consumers,
            )
        except KeyError as e:
            raise exception.InvalidManagedObjectDictError(field=str(e))



# Copyright (c) 2021 Mirantis Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Test cases for Vault key manager.
"""

import os
import tempfile

import requests_mock

from castellan.common import exception
from castellan.key_manager import vault_key_manager
from castellan.tests.unit.key_manager import test_key_manager


class VaultKeyManagerTestCase(test_key_manager.KeyManagerTestCase):
    def _create_key_manager(self):
        return vault_key_manager.VaultKeyManager(self.conf)

    def _create_sa_token_file(self, content="fake-sa-jwt-token"):
        """Create a temporary SA token file and register cleanup."""
        token_file = tempfile.NamedTemporaryFile(
            mode='w',
            suffix='.token',
            delete=False,
        )
        token_file.write(content)
        token_file.close()
        self.addCleanup(os.unlink, token_file.name)
        return token_file.name

    def _setup_token_auth(
        self,
        auth_method="kubernetes",
        role="my-role",
        token_content="fake-sa-jwt-token",
        auth_path=None,
    ):
        """Configure the key manager for token-based auth."""
        self.key_mgr._auth_method = auth_method
        self.key_mgr._token_role = role
        self.key_mgr._token_file = self._create_sa_token_file(token_content)
        self.key_mgr._auth_path = auth_path or auth_method

    def test_auth_headers_root_token(self):
        self.key_mgr._root_token_id = "spam"
        expected_headers = {"X-Vault-Token": "spam"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    def test_auth_headers_root_token_with_namespace(self):
        self.key_mgr._root_token_id = "spam"
        self.key_mgr._namespace = "ham"
        expected_headers = {
            "X-Vault-Token": "spam",
            "X-Vault-Namespace": "ham",
        }
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    @requests_mock.Mocker()
    def test_auth_headers_app_role(self, m):
        self.key_mgr._approle_role_id = "spam"
        self.key_mgr._approle_secret_id = "secret"
        m.post(
            "http://127.0.0.1:8200/v1/auth/approle/login",
            json={"auth": {"client_token": "token", "lease_duration": 3600}},
        )
        expected_headers = {"X-Vault-Token": "token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    @requests_mock.Mocker()
    def test_auth_headers_app_role_with_namespace(self, m):
        self.key_mgr._approle_role_id = "spam"
        self.key_mgr._approle_secret_id = "secret"
        self.key_mgr._namespace = "ham"
        m.post(
            "http://127.0.0.1:8200/v1/auth/approle/login",
            json={"auth": {"client_token": "token", "lease_duration": 3600}},
        )
        expected_headers = {
            "X-Vault-Token": "token",
            "X-Vault-Namespace": "ham",
        }
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    @requests_mock.Mocker()
    def test_auth_headers_app_role_token_caching(self, m):
        """Verify that the AppRole token is cached and reused."""
        self.key_mgr._approle_role_id = "spam"
        self.key_mgr._approle_secret_id = "secret"
        m.post(
            "http://127.0.0.1:8200/v1/auth/approle/login",
            json={"auth": {"client_token": "token", "lease_duration": 3600}},
        )
        headers1 = self.key_mgr._build_auth_headers()
        headers2 = self.key_mgr._build_auth_headers()
        self.assertEqual(headers1, headers2)
        self.assertEqual({"X-Vault-Token": "token"}, headers2)
        self.assertEqual(1, m.call_count)

    @requests_mock.Mocker()
    def test_auth_headers_token_file(self, m):
        self._setup_token_auth()
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes/login",
            json={
                "auth": {
                    "client_token": "k8s-token",
                    "lease_duration": 1800,
                },
            },
        )
        expected_headers = {"X-Vault-Token": "k8s-token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())
        # Verify the request body
        self.assertEqual(
            m.last_request.json(),
            {"role": "my-role", "jwt": "fake-sa-jwt-token"},
        )

    @requests_mock.Mocker()
    def test_auth_headers_token_file_custom_path(self, m):
        self._setup_token_auth(auth_method="jwt")
        m.post(
            "http://127.0.0.1:8200/v1/auth/jwt/login",
            json={
                "auth": {
                    "client_token": "jwt-token",
                    "lease_duration": 900,
                },
            },
        )
        expected_headers = {"X-Vault-Token": "jwt-token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    @requests_mock.Mocker()
    def test_auth_headers_token_file_with_namespace(self, m):
        self._setup_token_auth()
        self.key_mgr._namespace = "my-namespace"
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes/login",
            json={
                "auth": {
                    "client_token": "k8s-ns-token",
                    "lease_duration": 1800,
                },
            },
        )
        expected_headers = {
            "X-Vault-Token": "k8s-ns-token",
            "X-Vault-Namespace": "my-namespace",
        }
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    def test_auth_headers_token_file_missing(self):
        self.key_mgr._auth_method = 'kubernetes'
        self.key_mgr._token_role = "my-role"
        self.key_mgr._auth_path = "kubernetes"
        self.key_mgr._token_file = "/nonexistent/path/token"
        self.assertRaises(
            exception.KeyManagerError, self.key_mgr._build_auth_headers
        )

    @requests_mock.Mocker()
    def test_auth_headers_token_file_forbidden(self, m):
        self._setup_token_auth(role="bad-role")
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes/login",
            status_code=403,
        )
        self.assertRaises(
            exception.Forbidden, self.key_mgr._build_auth_headers
        )

    @requests_mock.Mocker()
    def test_auth_headers_token_file_bad_request(self, m):
        self._setup_token_auth()
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes/login",
            status_code=400,
            json={"errors": ["missing role", "invalid jwt"]},
        )
        self.assertRaises(
            exception.KeyManagerError, self.key_mgr._build_auth_headers
        )

    @requests_mock.Mocker()
    def test_auth_headers_token_file_caching(self, m):
        """Verify that the token from file-based auth is cached and reused."""
        self._setup_token_auth()
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes/login",
            json={
                "auth": {
                    "client_token": "cached-token",
                    "lease_duration": 3600,
                },
            },
        )
        # First call -- authenticates
        headers1 = self.key_mgr._build_auth_headers()
        # Second call -- should use cached token
        headers2 = self.key_mgr._build_auth_headers()
        self.assertEqual(headers1, headers2)
        self.assertEqual({"X-Vault-Token": "cached-token"}, headers2)
        # Only one HTTP request should have been made
        self.assertEqual(1, m.call_count)

    def test_auth_priority_root_over_token_file(self):
        """Root token takes priority over token file auth."""
        self.key_mgr._root_token_id = "root-token"
        self.key_mgr._auth_method = 'kubernetes'
        self.key_mgr._token_role = "my-role"
        expected_headers = {"X-Vault-Token": "root-token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    @requests_mock.Mocker()
    def test_auth_priority_root_over_approle(self, m):
        """Root token takes priority over AppRole auth."""
        self.key_mgr._root_token_id = "root-token"
        self.key_mgr._approle_role_id = "app-role"
        expected_headers = {"X-Vault-Token": "root-token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    def test_auth_token_missing_role(self):
        """auth_method=kubernetes without token_role raises error."""
        self.key_mgr._auth_method = 'kubernetes'
        self.key_mgr._auth_path = 'kubernetes'
        self.assertRaises(
            exception.KeyManagerError, self.key_mgr._build_auth_headers
        )

    @requests_mock.Mocker()
    def test_auth_headers_token_file_custom_auth_path(self, m):
        """auth_path overrides auth_method in the login URL."""
        self._setup_token_auth(
            auth_method="kubernetes",
            auth_path="kubernetes-my-custom-cluster",
        )
        m.post(
            "http://127.0.0.1:8200/v1/auth/kubernetes-my-custom-cluster/login",
            json={
                "auth": {
                    "client_token": "custom-token",
                    "lease_duration": 1800,
                },
            },
        )
        expected_headers = {"X-Vault-Token": "custom-token"}
        self.assertEqual(expected_headers, self.key_mgr._build_auth_headers())

    def test_auth_token_missing_token_file(self):
        """auth_method=kubernetes without token_file raises error."""
        self.key_mgr._auth_method = 'kubernetes'
        self.key_mgr._auth_path = 'kubernetes'
        self.key_mgr._token_role = "my-role"
        self.key_mgr._token_file = None
        self.assertRaises(
            exception.KeyManagerError, self.key_mgr._build_auth_headers
        )

    def test_auth_approle_no_role_id_returns_empty(self):
        """auth_method=approle without approle_role_id returns empty."""
        self.key_mgr._auth_method = 'approle'
        self.assertEqual({}, self.key_mgr._build_auth_headers())
